%% tclp__cache.pl
%% library for caching informations
%% Copyright (C) 2004 INRIA Emmanuel Coquery

%% This library is free software; you can redistribute it and/or
%% modify it under the terms of the GNU Lesser General Public
%% License as published by the Free Software Foundation; either
%% version 2.1 of the License, or (at your option) any later version.

%% This library is distributed in the hope that it will be useful,
%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
%% Lesser General Public License for more details.

%% You should have received a copy of the GNU Lesser General Public
%% License along with this library; if not, write to the Free Software
%% Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%                                                                           %%
%%                                   TCLP                                    %%
%%                                                                           %%
%%         Emmanuel Coquery, projet Contraintes, INRIA Rocquencourt          %%
%%                                                                           %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% $Header: /home/yquem/contraintes/coquery/cvs/contraintes/tclpchr/src/tclp/tclp__cache.pl,v 1.1 2004/11/05 18:53:49 manu Exp $

:- module(tclp__cache, [ cache_trigger/2,
                         load_cache/1,
                         store_cache/1,
                         put_cache_info/3,
                         restitute_cache_info/1,
                         flush_cache_info/1,
                         cache_is_valid/1]).

:- use_module(library(system)).

%% dynamic predicates containing cache triggers to be checked
%% dynamic predicates containing cache triggers to be saved
:- dynamic old_cache_triggers/2.
:- dynamic new_cache_triggers/2.

%% dynamic predicate telling is the cache is valid.
%% false by default
:- dynamic cache_is_valid/1.

%% sets the cache as invalid
invalidate_cache(Cache) :- retractall(cache_is_valid(Cache)).

%% sets the cache as valid, that is if it has not been invalidated
%% and no old triggers for the cache remains
validate_cache(Cache) :-
        assert((cache_is_valid(Cache):-!,\+ old_cache_triggers(Cache,_))).

%% cache_trigger(Cache,T) says that the term T is a trigger for the cache Cache:
%% when calling cache_is_valid, the cache triggers must correspond exactly
%% to the ones registered in the cache file.
cache_trigger(Cache,T) :-
        (old_cache_triggers(Cache,T) ->
            retract(old_cache_triggers(Cache,T))
        ;
            invalidate_cache(Cache)),
        assert(new_cache_triggers(Cache,T)).

%% dynamic predicate containing cache information
:- dynamic cache_info/3.

%% put_cache_info(Cache,M,C)
%% add information into the cache Cache in the form of a clause. The clause is
%% asserted in the module M
put_cache_info(Cache,M,C) :-
        assert(cache_info(Cache,M,C)).

%% restitute_cache_info(Cache)
%% restitutes all the information from the cache Cache
%% by, for each clause cache_info(Cache,M,C), asserting C in the module M
restitute_cache_info(Cache) :-
        ( 
          cache_info(Cache,M,C),
          M:assert(C),
          fail
        ;
          true
        ),!.

%% name of the cache file
cache_file(Cache,Cache) :-
        atom_concat(_,'.tclp_cache',Cache),!.
cache_file(Cache,File) :-
        atom_concat(Cache,'.tclp_cache',File).

%% store cache information for the cache Cache
store_cache(Cache) :-
        cache_file(Cache,File),
        open(File,write,Stream),
        !,
        (
          new_cache_triggers(Cache,T),
          write_canonical(Stream,old_cache_triggers(Cache,T)),
          write(Stream,'.'),nl(Stream),
          fail
        ;
          cache_info(Cache,M,C),
          write_canonical(Stream,cache_info(Cache,M,C)),
          write(Stream,'.'),nl(Stream),
          fail
        ;
          true
        ),
        close(Stream),
        !.

%% load information from a cache file
load_cache(Cache) :-
        cache_file(Cache,File),
        file_exists(File),
        !,
        open(File,read,Stream),
        !,
        repeat,
        read(Stream,Term),
        (
          Term == end_of_file
        ;
          assert(Term),
          fail
        ),
        close(Stream),
        validate_cache(Cache),
        !.
load_cache(Cache) :-
                                % in case the file doesn't exists
        invalidate_cache(Cache).

%% removes all information about some cache
flush_cache_info(Cache) :-
        retractall(cache_info(Cache,_,_)).
