%% Library for dealing with type aliases
%% Copyright (C) 2005 INRIA Emmanuel Coquery

%% This library is free software; you can redistribute it and/or
%% modify it under the terms of the GNU Lesser General Public
%% License as published by the Free Software Foundation; either
%% version 2.1 of the License, or (at your option) any later version.

%% This library is distributed in the hope that it will be useful,
%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
%% Lesser General Public License for more details.

%% You should have received a copy of the GNU Lesser General Public
%% License along with this library; if not, write to the Free Software
%% Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%                                                                           %%
%%                                   TCLP                                    %%
%%                                                                           %%
%%         Emmanuel Coquery, projet Contraintes, INRIA Rocquencourt          %%
%%                                                                           %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% $Header: /home/yquem/contraintes/coquery/cvs/contraintes/tclpchr/src/tclp/tclp__type_aliases.pl,v 1.3 2006/07/17 14:51:29 ecoquery Exp $

:- module(tclp__type_aliases,[ type_alias/2,
                               add_alias/3,
                               get_aliases_in_term/2,
                               print_aliases/2,
                               print_aliases/1,
                               reinit_aliases_goals/0,
                               unalias_terms/2,
                               unalias_term/2
                             ]).

:- use_module(tclp(tclp__utils)).
:- use_module(tclp(tclp__reader)).
:- use_module(tclp(tclp__printer)).
:- use_module(tclp(tclp__errors)).
:- use_module(library(lists)).
:- use_module(tclp(tclp__handle_goals)).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Handling the dynamic predicate type_alias/2
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% type_alias(A,T) is true if A is an alias for T
:- dynamic type_alias/2.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Adds a new alias and check the correcness of the declaration.
add_alias(Alias,_,TermData) :-
        var(Alias),!,
        print_error(TermData,['An alias can not be a variable']).
add_alias(Alias,_,TermData) :-
        \+ args_are_var(Alias),!,
        print_error(TermData,['An alias can have only variable arguments in its declaration']).
add_alias(Alias,Type,_) :-
        type_alias(Alias,Type),!.
add_alias(Alias,Type,_TermData) :-
        assert(type_alias(Alias,Type)),
        verbose(actions,['added alias: ',Alias,' = ',Type]).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Getting/printing aliases from terms
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% get aliases from terms
get_aliases_list([],[]).
get_aliases_list([X|L],Res) :-
        get_aliases(X,XA),
        get_aliases_list(L,LA),
        append(XA,LA,Res).

get_aliases(X,[]) :-
        var(X),!.
get_aliases(T,Res) :-
        findall(T-A,type_alias(A,T),LTA),
        T =.. [_|Args],
        get_aliases_list(Args,LArgsA),
        append(LTA,LArgsA,Res).

get_aliases_in_term(X,LA) :-
        get_aliases(X,UnsLA),
        sort(UnsLA,LA).
%        sort_and_unify(UnsLA,LA).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Transform a term into (one of) its aliased versions
aliased_term(T,T) :-
        var(T),!.
aliased_term(T,A) :-
        (type_alias(T,B) ; T=B),
        B =.. [ H | BArgs ],
        aliased_terms(BArgs, AArgs),
        A =.. [ H | AArgs ].
aliased_terms([],[]).
aliased_terms([X1|R1],[X2|R2]) :-
        aliased_term(X1,X2),
        aliased_terms(R1,R2).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Transforms a type that uses aliases into a type without alias
unalias_term(A,A) :-
        var(A),!.
unalias_term(Alias,Type) :-
        Alias =.. [ Head | AliasArgs ],
        unalias_terms(AliasArgs, UnaliasedArgs),
        PartiallyUnaliased =.. [ Head | UnaliasedArgs ],
        ( type_alias(PartiallyUnaliased,Type)
        -> true
        ; PartiallyUnaliased = Type).

unalias_terms([],[]).
unalias_terms([A|RA],[T|RT]) :-
        unalias_term(A,T),
        unalias_terms(RA,RT).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Print aliases occuring in a term
print_aliases(Term, Printer) :-
        get_aliases_in_term(Term,LA),
        (LA = [] -> true; pprint(Printer,['Type aliases:'])),
        print_alias_pairs(LA,Printer).

print_aliases(Term) :-
        print_aliases(Term,printer('! ')).

print_alias_pairs([],_).
print_alias_pairs([A-T|L],Printer) :-
        pprint(Printer,[ A,' = ',T]),
        print_alias_pairs(L,Printer).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Enable alias declarations
:- op(999,fx,typealias).

handle_alias((typealias A = B), TermData) :- !,
        add_alias(A,B,TermData).
handle_alias((typealias _),TermData) :-
        print_error(TermData,['A type alias declaration is of the form: :- typealias t(A,B,C,...) = f(x,y,...), where t is the type alias, f is a type constructor and x and y are types.']).

typealias _.

reinit_aliases_goals :-
        retractall(type_alias(_,_)),
        tclp__executable(typealias(_)),
        tclp__add_hook(typealias(A),TData1,tclp__type_aliases:handle_alias(typealias(A),TData1)).

:- reinit_aliases_goals.
